import maya.OpenMaya as OpenMaya1


class SuppressScriptEditorOutput:
    """
    Suppresses output to the scriptEditor
    """

    def __init__(self):
        self.callback_id = OpenMaya1.MCommandMessage.addCommandOutputFilterCallback(
            self.suppress
        )

    @staticmethod
    def suppress(_, __, filter_output, ___):
        """
        This is the callback function that gets called when Maya wants to
        print something suppressing the print
        """
        OpenMaya1.MScriptUtil.setBool(filter_output, True)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        OpenMaya1.MMessage.removeCallback(self.callback_id)



import pytest
from maya_brew import OpenMaya2, cmds
from maya_brew.attributes.node_attribute import (
    Attribute,
    FloatAttribute,
    MessageAttribute,
)
from maya_brew.log import get_logger
from maya_brew.nodes.node_types import DagNode, Node

logger = get_logger(__name__)


def test_cast_attributes(brew_transform):
    tx = Attribute(f"{brew_transform.node_path}.translateX")
    assert isinstance(tx, FloatAttribute)
    message = Attribute(f"{brew_transform}.message")
    assert isinstance(message, MessageAttribute)


def test_message_attribute_get_value_error(brew_transform):
    with pytest.raises(AttributeError):
        Attribute._get_value(brew_transform, "message")


def test_attribute_get_value(brew_transform):
    tx = Attribute(f"{brew_transform.node_path}.translateX")
    value = tx.get()
    assert value == 0
    cmds.setAttr(f"{brew_transform.node_path}.translateX", 5)
    value = FloatAttribute._get_value(brew_transform, "translateX")
    assert value == 5


def test_factory_with_mplug(translatex_plug):
    a = Attribute(translatex_plug)
    assert isinstance(a, FloatAttribute)
    assert a.get() == 0


def test_subclass_with_mplug(translatex_plug):
    fa = FloatAttribute(translatex_plug)
    assert fa.get() == 0


def test_invalid_path_missing_dot():
    with pytest.raises(ValueError):
        Attribute("bad_path_without_dot")


def test_invalid_plug_type():
    with pytest.raises(ValueError):
        Attribute(123)  # type: ignore


def test_get_node_from_plug_dag_and_non_dag(translatex_plug, non_dag_plug):
    dag = Attribute._get_node_from_plug(translatex_plug)
    assert isinstance(dag, DagNode)
    expected_dag_path = OpenMaya2.MFnDagNode(translatex_plug.node()).fullPathName()
    assert str(dag) == expected_dag_path

    non_dag = Attribute._get_node_from_plug(non_dag_plug)
    assert isinstance(non_dag, Node)
    assert not isinstance(non_dag, DagNode)
    expected_non_dag_name = OpenMaya2.MFnDependencyNode(non_dag_plug.node()).name()
    non_dag_name = str(non_dag)
    assert non_dag_name == expected_non_dag_name


def test_attribute_set_instance(brew_transform):
    tx = Attribute(f"{brew_transform.node_path}.translateX")
    tx.set(3.14)
    assert cmds.getAttr(f"{brew_transform.node_path}.translateX") == 3.14


def test_attribute_set_redo(brew_transform):
    tx = Attribute(f"{brew_transform.node_path}.translateX")
    initial = cmds.getAttr(f"{brew_transform.node_path}.translateX")
    tx.set(8.88)
    with logger.silence():
        cmds.undo()
    assert cmds.getAttr(f"{brew_transform.node_path}.translateX") == initial
    with logger.silence():
        cmds.redo()
    assert cmds.getAttr(f"{brew_transform.node_path}.translateX") == 8.88


def test_message_attribute_set_value_error(brew_transform):
    message_attr = Attribute(f"{brew_transform}.message")
    with pytest.raises(AttributeError):
        message_attr.set(1)


def test_connect_and_force_overwrite_and_disconnect(brew_transform, test_cube):
    tx_attr = Attribute(f"{brew_transform.node_path}.translateX")
    ty_attr = Attribute(f"{brew_transform.node_path}.translateY")
    cube_ty = Attribute(f"{test_cube}.translateY")

    def get_ty_connections():
        return cmds.listConnections(cube_ty.plug.name(), source=True, plugs=True)

    tx_attr.connect(cube_ty)
    connections = get_ty_connections()
    expected_connection = tx_attr.plug.name()
    assert connections == [expected_connection]

    cmds.undo()
    assert not get_ty_connections()
    cmds.redo()
    assert get_ty_connections() == [expected_connection]

    ty_attr.connect(cube_ty, force=True)
    connections = get_ty_connections()
    expected_connection = ty_attr.plug.name()
    assert connections == [expected_connection]

    cmds.undo()
    assert get_ty_connections() == [tx_attr.plug.name()]
    cmds.redo()
    assert get_ty_connections() == [expected_connection]

    ty_attr.disconnect(cube_ty)
    assert not get_ty_connections()

    cmds.undo()
    assert get_ty_connections() == [ty_attr.plug.name()]
    cmds.redo()
    assert not get_ty_connections()



import maya_brew
import maya_brew.nodes.node_types
import maya_brew.scene
import pytest
from maya_brew import OpenMaya2


@pytest.fixture()
def test_namespace():
    return "my_namespace"


@pytest.fixture()
def test_cube_short_name():
    return "test_cube"


@pytest.fixture()
def test_cube(test_namespace, test_cube_short_name):
    cube, history = maya_brew.cmds.polyCube(
        name=f"{test_namespace}:{test_cube_short_name}"
    )
    yield cube


@pytest.fixture()
def brew_transform():
    """
    Fixture to return the node object of the test cube.
    """
    yield maya_brew.nodes.node_types.Transform.create("test_transform")


@pytest.fixture(autouse=True)
def new_scene():
    return maya_brew.scene.new_file()


@pytest.fixture()
def translatex_plug(brew_transform):
    """Provide the MPlug for translateX on the empty_transform."""
    fn = brew_transform.get_mfndependency_node()
    return fn.findPlug("translateX", False)


@pytest.fixture()
def non_dag_plug():
    """Provide the MPlug for a non-DAG dependency node (multiplyDivide.input1X)."""
    name = "test_multiplyDivide"
    maya_brew.cmds.createNode("multiplyDivide", name=name)
    sel = OpenMaya2.MSelectionList()
    sel.add(name)
    mobj = sel.getDependNode(0)
    fn = OpenMaya2.MFnDependencyNode(mobj)
    return fn.findPlug("input1X", False)



from .. import log
from .interactive_utils import SuppressScriptEditorOutput

# note this logger is defined before SilenceMayaContextManager is added so will not
# silence script editor output
logger = log.get_logger(__name__)
# logger.setLevel("DEBUG")


class SilenceMayaContextManager(
    SuppressScriptEditorOutput, log.LogAllLevelsContextManager
):
    pass


def initialize_maya():
    log.SilenceContextManager = SilenceMayaContextManager
    logger.debug("Initializing interactive maya...")



import importlib
import sys
from pathlib import Path

from .log import get_logger

logger = get_logger(__name__)
logger.setLevel("DEBUG")


def reload_mb():
    """
    Workaround to reload entire maya-brew library.
    Removes all maya-brew modules from 'sys.modules' and imports it again
    """

    import maya_brew

    mb_path = Path(maya_brew.__file__)
    mb_modules = list()
    for m in [p for p in sys.modules if p.startswith(mb_path.parent.name)]:
        mb_modules.append(m)
        sys.modules.pop(m)
    for m in mb_modules:
        logger.debug(f"re-importing {m}")
        importlib.import_module(m)



import maya_brew
import maya_brew.nodes.cast
import pytest


def test_string_to_dag_path(test_cube, test_namespace, test_cube_short_name):
    cube_full_path = maya_brew.cmds.ls(test_cube, long=True)[0]
    assert cube_full_path == f"|{test_namespace}:{test_cube_short_name}"
    dag_path = maya_brew.nodes.cast.get_dag_path_from_string(cube_full_path)
    assert isinstance(dag_path, maya_brew.OpenMaya2.MDagPath)


def test_invalid_string_to_dag_path(test_cube, test_cube_short_name):
    with pytest.raises(maya_brew.nodes.cast.NonExistingDagPath):
        maya_brew.nodes.cast.get_dag_path_from_string("invalid_name")
    with pytest.raises(maya_brew.nodes.cast.NonExistingDagPath):
        maya_brew.nodes.cast.get_dag_path_from_string(test_cube_short_name)
    with pytest.raises(maya_brew.nodes.cast.InvalidDagPath):
        maya_brew.nodes.cast.get_dag_path_from_string("lambert1")


def test_get_long_name_from_maya_string(
    test_cube, test_namespace, test_cube_short_name
):
    cube_full_name = f"{test_namespace}:{test_cube_short_name}"
    cube_full_path = maya_brew.nodes.cast.get_long_name_from_maya_string(cube_full_name)
    assert cube_full_path == f"|{cube_full_name}"
    grouped_cube = maya_brew.cmds.polyCube(name=test_cube_short_name)[0]
    group_name = "my_group"
    maya_brew.cmds.group(name=group_name, empty=True)
    maya_brew.cmds.parent(grouped_cube, group_name)
    grouped_cube_full_path = maya_brew.nodes.cast.get_long_name_from_maya_string(
        grouped_cube
    )
    assert group_name in grouped_cube_full_path
    new_cube_with_same_short_name = maya_brew.cmds.polyCube(name=test_cube_short_name)[
        0
    ]
    # we can cast it from maya's return value
    new_cube_full_path = maya_brew.nodes.cast.get_long_name_from_maya_string(
        new_cube_with_same_short_name
    )
    # if we try to cast it based on short name we should get two matches
    with pytest.raises(maya_brew.nodes.cast.MultipleMatchingNodes):
        maya_brew.nodes.cast.get_long_name_from_maya_string(test_cube_short_name)
    assert test_namespace not in new_cube_full_path
    with pytest.raises(maya_brew.nodes.cast.NoMatchingNodes):
        maya_brew.nodes.cast.get_long_name_from_maya_string("invalid_name")



from typing import Any, Callable, Dict, Self

from .. import OpenMaya2, cmds
from ..log import get_logger
from ..nodes import cast

logger = get_logger(__name__)
logger.setLevel("DEBUG")


class Node:
    _cmds_creator: Callable[..., str]
    _cmds_creator_args: Dict[str, Any] = {}

    def __init__(self, node_path: str):
        self.node_path = node_path

    def __str__(self):
        return self.node_path

    def __repr__(self):
        return f'{type(self).__name__}("{str(self)}")'

    def rename(self, new_name: str, *args, **kwargs) -> str:
        """
        Rename the current node. If new name is not unique maya will resolve a new unique name.
        :param new_name: The new name of the node.
        :return: The new name of the node.
        """
        actual_new_name = cmds.rename(str(self), new_name, *args, **kwargs)
        self.node_path = actual_new_name
        return actual_new_name

    def _cmd_caller(self, func_name, *args, **kwargs):
        return cmds.attr(func_name, self.node_path, *args, **kwargs)

    @classmethod
    def create(cls, name: str, **kwargs) -> Self:
        """
        Create a new transform node.
        :param name: The name of the new transform node.
        :return: The new transform node.
        """
        kwargs.update(cls._cmds_creator_args)
        kwargs["name"] = name
        cmds_value = cls._cmds_creator(**kwargs)  # noqa: E1102
        if cmds_value != name:
            logger.debug(f"Name was not unique. New name: {cmds_value}.")
        return cls(cmds_value)


class DagNode(Node):
    def __init__(self, node_path: str):
        super().__init__(node_path)
        self.dag_path = cast.get_dag_path_from_string(node_path)

    def __str__(self):
        return self.get_full_path()

    def get_depend_node(self) -> OpenMaya2.MObject:
        """
        Get the depend node of the current node.
        :return: The depend node of the current node.
        """
        return self.dag_path.node()

    def get_full_path(self) -> str:
        """
        Get the full path of the current node.
        :return: The full path of the current node.
        """
        full_path = self.dag_path.fullPathName()
        if not full_path:
            logger.warning(f"Node has been deleted. Original path: {self.node_path}")
        return full_path

    def get_mfndependency_node(self) -> OpenMaya2.MFnDependencyNode:
        """
        Get the MFnDependencyNode of the current node.
        :return: The MFnDependencyNode of the current node.
        """
        return OpenMaya2.MFnDependencyNode(self.get_depend_node())


class Transform(DagNode):
    _cmds_creator = cmds.group
    _cmds_creator_args = {"empty": True}



import maya_brew

from .exceptions import MayaBrewException


class SceneException(MayaBrewException):
    """Base class for scene-related exceptions."""

    pass


class UnsavedChanges(SceneException, RuntimeError):
    """Exception raised when there are unsaved changes in the scene."""

    pass


def new_file(force=True):
    try:
        return maya_brew.cmds.file(newFile=True, force=force)
    except RuntimeError as e:
        message = str(e).strip()
        if message == "Unsaved changes.":
            raise UnsavedChanges(message) from e
        raise SceneException(message) from e



"""
Utilities for setting up logging in modules
"""

import datetime
import fnmatch
import logging
import os
import sys
import threading
import time
from typing import Callable, List, Optional, Union

import structlog

from . import PACKAGE_NAME, get_bool_env_variable

PROPAGATE = get_bool_env_variable(f"{PACKAGE_NAME}_LOG_PROPAGATE", False)


TELEMETRY_PROCESSORS: List[Callable] = []


def telemetry_runner(target_logger, method_name, event_dict):
    """
    Runs processors for telemetry events or events where logger is enabled for event level.
    Drops telemetry events after processing
    """
    is_telemetry_event = event_dict.get("event_type") == "telemetry"
    if is_telemetry_event:
        run_telemetry_processors(target_logger, method_name, event_dict)
        raise structlog.DropEvent
    event_level = event_dict.get("level")
    numeric_level = getattr(logging, event_level.upper(), None)
    if target_logger.isEnabledFor(numeric_level):
        run_telemetry_processors(target_logger, method_name, event_dict)
    return event_dict


def run_telemetry_processors(target_logger, method_name, event_dict):
    for processor in TELEMETRY_PROCESSORS:
        processor(target_logger, method_name, event_dict)
        return event_dict


def format_console_time(target_logger, method_name, event_dict):
    event_dict["timestamp"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return event_dict


def configure(use_colors=False):
    # maya doesn't seem to play nice with colors, and once structlog is configured to use colors it seems to
    # cause problems when reconfiguring without

    if not structlog.is_configured():
        logging.basicConfig(level=logging.INFO)
        logging.Logger.propagate = bool(PROPAGATE)

    structlog.configure(
        processors=[
            structlog.stdlib.add_log_level,
            structlog.contextvars.merge_contextvars,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso", utc=False),
            structlog.processors.CallsiteParameterAdder(
                {
                    structlog.processors.CallsiteParameter.FILENAME,
                    structlog.processors.CallsiteParameter.PATHNAME,
                    structlog.processors.CallsiteParameter.MODULE,
                    structlog.processors.CallsiteParameter.FUNC_NAME,
                    structlog.processors.CallsiteParameter.LINENO,
                }
            ),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.UnicodeDecoder(),
            telemetry_runner,
            format_console_time,
            structlog.dev.ConsoleRenderer(colors=use_colors),
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


class LogLevelContextManager:
    def __init__(self, target_logger, level):
        self.logger = target_logger
        self.level = level
        self.previous_level = target_logger.getEffectiveLevel()

    def __enter__(self):
        self.logger.setLevel(self.level)

    def __call__(self, func):
        """Allow class to be used as decorator"""

        def decorator(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorator

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.setLevel(self.previous_level)


class LogAllLevelsContextManager:
    def __init__(self, min_level=logging.CRITICAL):
        self.min_level = min_level

    def __enter__(self):
        logging.disable(self.min_level)

    def __call__(self, func):
        """Allow class to be used as decorator"""

        def decorator(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorator

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


class SuppressStdOutStdErr:
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    """

    def __init__(
        self, filter_list: Optional[Union[List[str], str]] = None, enable=True
    ) -> None:
        self.enable = enable
        if self.enable:
            if filter_list is None:
                self.filter_words = []
            elif isinstance(filter_list, str):
                self.filter_words = [filter_list]
            else:
                self.filter_words = filter_list

    def __enter__(self):
        if self.enable:
            self._stdout = sys.stdout
            self._stderr = sys.stderr
            self._stdout_pipe = os.pipe()
            self._stderr_pipe = os.pipe()
            self._stdout_fd = os.dup(1)
            self._stderr_fd = os.dup(2)
            os.dup2(self._stdout_pipe[1], 1)
            os.dup2(self._stderr_pipe[1], 2)
            self._stdout_thread = self._start_thread(self._stdout_pipe[0], self._stdout)
            self._stderr_thread = self._start_thread(self._stderr_pipe[0], self._stderr)
            os.close(self._stdout_pipe[1])
            os.close(self._stderr_pipe[1])
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.enable:
            os.dup2(self._stdout_fd, 1)
            os.dup2(self._stderr_fd, 2)

            os.close(self._stdout_fd)
            os.close(self._stderr_fd)

            if self._stdout_thread.is_alive():
                self._stdout_thread.join()
            if self._stderr_thread.is_alive():
                self._stderr_thread.join()

    def _should_filter(self, line) -> bool:
        if not self.filter_words:
            return True
        line = line.strip()
        for pattern in self.filter_words:
            pattern = pattern
            if "*" in pattern:
                if fnmatch.fnmatch(line, pattern):
                    return True
            else:
                if pattern in line:
                    return True
        return False

    def _start_thread(self, read_fd, write_stream):
        def _filter_output():
            read_file = os.fdopen(read_fd)
            try:
                for line in read_file:
                    if not self._should_filter(line):
                        write_stream.write(line)
                        write_stream.flush()
            finally:
                read_file.close()

        thread = threading.Thread(target=_filter_output)
        thread.daemon = True
        thread.start()
        return thread


class ExecutionTimer:
    def __init__(self):
        self.elapsed_time = None

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, type, value, traceback):
        end_time = time.time()
        self.elapsed_time = end_time - self.start_time


class SilenceContextManager(SuppressStdOutStdErr):
    pass


class SingleLevelFilter(logging.Filter):
    def __init__(self, passlevel, reject):
        super(SingleLevelFilter, self).__init__()
        self.passlevel = passlevel
        self.reject = reject

    def filter(self, record):
        if self.reject:
            return record.levelno != self.passlevel
        else:
            return record.levelno == self.passlevel


class CustomLogger(structlog.stdlib.BoundLogger):
    timer: ExecutionTimer
    info_handler: logging.StreamHandler

    @staticmethod
    def all_log_levels(min_level=logging.CRITICAL):
        return LogAllLevelsContextManager(min_level)

    @staticmethod
    def silence(filter_list: Optional[List[str]] = None, enable=True):
        return SilenceContextManager(filter_list=filter_list, enable=enable)

    @staticmethod
    def at_level(target_logger, level):
        return LogLevelContextManager(target_logger, level)

    @staticmethod
    def info_to_stdout(target_logger=None):
        h1 = logging.StreamHandler(sys.stdout)
        f1 = SingleLevelFilter(logging.INFO, False)
        h1.addFilter(f1)
        target_logger.addHandler(h1)

    @staticmethod
    def send_telemetry(*args, **kwargs):
        """Sends message to any telemetry handlers on logger"""
        raise NotImplementedError

    def setLevel(self, level: Union[int, str]) -> None:
        super().setLevel(level)  # type: ignore[arg-type]


class TelemetryHandler:
    """Abstract class for logging handlers to send telemetry data"""

    pass


def get_logger(name) -> CustomLogger:
    """
    Simple stream-logger that can also be used as a context manager.
    Be aware of the hacky adding of methods to existing logger,
    but overriding return type to get correct inspection in IDEs
    """
    logger = structlog.stdlib.get_logger(name)  # noqa

    # Hacks since current versions of pytest doesn't allow subclassing
    # of logging.getLogger
    setattr(logger, "all_log_levels", CustomLogger.all_log_levels)

    def _silence(*args, **kwargs):
        return CustomLogger.silence(*args, **kwargs)  # noqa

    setattr(logger, "silence", _silence)

    setattr(logger, "timer", ExecutionTimer)

    def _send_telemetry(*args, **kwargs):
        kwargs["event_type"] = "telemetry"
        logger.info(*args, **kwargs)

    setattr(logger, "send_telemetry", _send_telemetry)

    def _info_to_stdout():
        return CustomLogger.info_to_stdout(logger)

    setattr(logger, "info_to_stdout", _info_to_stdout)

    def at_level(level):
        return CustomLogger.at_level(logger, level)

    setattr(logger, "at_level", at_level)
    return logger  # type: ignore[return-value]


def mock_telemetry_processor(logger, method_name, event_dict):
    print(f"Sending to some telemetry service: {event_dict}")
    return event_dict


configure()

if __name__ == "__main__":
    TELEMETRY_PROCESSORS.append(mock_telemetry_processor)
    configure(use_colors=True)
    example_logger = get_logger(__name__)
    example_logger.info("Logging configured")
    example_logger.debug(
        "This is a debug message and shouldn't be displayed as logging level is set to info"
    )
    example_logger.setLevel("DEBUG")
    with example_logger.silence(filter_list=["filtered"]):
        example_logger.info("this should be filtered out")
        example_logger.info("this should not be silenced!")
    example_logger.debug(
        "This is a debug message and should be displayed as logging level is set to debug"
    )



import logging

import maya_brew.nodes.cast
import maya_brew.nodes.node_types


def test_dag_node_string(test_cube, caplog):
    dag_node = maya_brew.nodes.node_types.DagNode(test_cube)
    full_node_path = maya_brew.nodes.cast.get_long_name_from_maya_string(test_cube)
    assert str(dag_node) == full_node_path
    # rename to see warning display correct node path when deleted
    new_name = "new_name"
    dag_node.rename(new_name)
    maya_brew.cmds.delete(str(dag_node))
    with caplog.at_level(logging.WARNING):
        str(dag_node)
        node_path = dag_node.node_path
        assert node_path in str(caplog.text)


def test_dag_node_rename(test_cube, caplog):
    dag_node = maya_brew.nodes.node_types.DagNode(test_cube)
    new_name = "new_name"
    new_resolved_name = dag_node.rename(new_name)
    assert new_resolved_name == new_name
    another_cube = maya_brew.cmds.polyCube(name="another_cube")[0]
    another_dag_node = maya_brew.nodes.node_types.DagNode(another_cube)
    another_cube_new_name = another_dag_node.rename(new_name)
    assert another_cube_new_name != new_name
    assert another_cube_new_name == "new_name1"
    another_dag_node_full_path = f"|{another_cube_new_name}"
    assert str(another_dag_node) == another_dag_node_full_path
    assert another_dag_node.dag_path.fullPathName() == another_dag_node_full_path


def test_Transform_repr():
    name = "hello"
    dag_node = maya_brew.nodes.node_types.Transform.create(name)
    repr = dag_node.__repr__()
    assert name in repr
    assert dag_node.get_full_path() in repr
    assert type(dag_node).__name__ in repr



import maya_brew
import maya_brew.scene
import pytest


def test_new_scene():
    cube, history = maya_brew.cmds.polyCube()
    with pytest.raises(maya_brew.scene.UnsavedChanges):
        maya_brew.scene.new_file(force=False)
    maya_brew.scene.new_file(force=True)
    assert maya_brew.cmds.objExists(cube) is False


def test_UnsavedChanges_is_RuntimeError():
    try:
        raise maya_brew.scene.UnsavedChanges("some test error")
    except RuntimeError as e:
        assert isinstance(e, maya_brew.scene.UnsavedChanges)



from maya_brew.maya_initializer import standalone_init


def test_standalone_can_use_cmds():
    standalone_init.initialize_maya()

    import maya.cmds as cmds

    (cube,) = cmds.polyCube(constructionHistory=False)
    assert isinstance(cube, str)


def test_standalone_not_interactive():
    from maya_brew.maya_initializer.shared import is_interactive_maya

    assert not is_interactive_maya()



import logging
import os

import maya.api._OpenMaya_py2 as OpenMaya2  # noqa
import maya.cmds as cmds  # noqa: F401

PACKAGE_NAME = "MAYA_BREW"


def get_bool_env_variable(name: str, default=True):
    value = os.getenv(name)
    if value is None:
        return default
    return value.lower() in ("true", "1", "t", "yes", "y")


MAYA_BREW_AUTOINIT = get_bool_env_variable(f"{PACKAGE_NAME}_AUTOINIT", True)


def init():
    old_handlers = logging.root.handlers[:]
    old_log_level = logging.getLogger().level
    import maya_brew.log
    from maya_brew.maya_initializer.shared import is_interactive_maya

    interactive = is_interactive_maya()
    if interactive:
        maya_brew.log.configure(use_colors=False)
    else:
        maya_brew.log.configure(
            use_colors=get_bool_env_variable(f"{PACKAGE_NAME}_USE_LOG_COLORS", True)
        )
    logger = maya_brew.log.get_logger(__name__)
    logger.info("Initializing maya...")
    os.environ["MAYA_NO_WARNING_FOR_MISSING_DEFAULT_RENDERER"] = "1"
    if is_interactive_maya():
        from maya_brew.maya_initializer import interactive_init

        interactive_init.initialize_maya()
    else:
        from maya_brew.maya_initializer import standalone_init

        standalone_init.initialize_maya()
    logging.root.handlers = old_handlers
    logging.getLogger().setLevel(old_log_level)
    logging.debug("maya initialized")


if MAYA_BREW_AUTOINIT:
    init()



from .. import OpenMaya2, cmds, exceptions


class NodeCastingException(exceptions.MayaBrewException):
    """Base class for node casting exceptions."""

    pass


class InvalidDagPath(NodeCastingException, RuntimeError, ValueError, TypeError):
    """Exception raised when a DAG path is invalid."""

    pass


class NonExistingDagPath(NodeCastingException, RuntimeError, ValueError, TypeError):
    """Exception raised when a DAG path is invalid."""

    pass


class MultipleMatchingNodes(NodeCastingException):
    """Exception raised when multiple nodes match a given name."""

    pass


class NoMatchingNodes(NodeCastingException):
    """Exception raised when no nodes match a given name."""

    pass


def get_dag_path_from_string(node_path: str) -> OpenMaya2.MDagPath:
    """
    Convert a string representation of a node path to an OpenMaya2.MDagPath object.
    """
    selection_list = OpenMaya2.MSelectionList()
    try:
        selection_list.add(node_path)
        dag_path = selection_list.getDagPath(0)
    except RuntimeError as e:
        if "Object does not exist" in str(e):
            raise NonExistingDagPath(f"Dag Path does not exist: {node_path}") from e
        raise
    except TypeError as e:
        if "item is not a DAG path" in str(e):
            raise InvalidDagPath(f"{node_path} exists but is not a dag object") from e
        raise
    return dag_path


def get_long_name_from_maya_string(name: str) -> str:
    """
    Convert a unique maya.cmds node string to a long name.
    """
    matching_nodes = cmds.ls(name, long=True)
    if not matching_nodes:
        raise NoMatchingNodes(f"No objects matching '{name}' in scene")
    num_nodes = len(matching_nodes)
    if num_nodes > 1:
        raise MultipleMatchingNodes(
            f"Multiple objects matching '{name}' in scene. Found {num_nodes} nodes: {matching_nodes}"
        )
    return matching_nodes[0]



from ..log import get_logger

logger = get_logger(__name__)
# logger.setLevel("DEBUG")


def initialize_maya():
    if cmds_about_exists():
        return
    with logger.silence(["This plugin does not support createPlatformOpenGLContext!"]):
        import maya.standalone

        maya.standalone.initialize(name="initialized_by_maya_brew")
    logger.debug("initialized maya standalone")


def cmds_about_exists():
    try:
        from maya.cmds import about  # noqa: F401

        return True
    except ImportError:
        return False


if __name__ == "__main__":
    initialize_maya()



import typing

from .. import OpenMaya2, cmds
from ..nodes import cast
from ..nodes.node_types import DagNode, Node

PlugInput = typing.Union[OpenMaya2.MPlug, str]
A = typing.TypeVar("A", bound="Attribute")


class Attribute:
    _getter_type: str

    def __new__(cls: type[A], plug_or_path: PlugInput) -> A:
        if cls is not Attribute:
            return typing.cast(A, super().__new__(cls))

        if isinstance(plug_or_path, OpenMaya2.MPlug):
            plug = plug_or_path
        elif isinstance(plug_or_path, str):
            plug = cls._plug_from_path(plug_or_path)
        else:
            raise ValueError("plug_or_path must be an MPlug or attribute path string.")

        try:
            mobj_attr = plug.attribute()
        except (RuntimeError, ValueError) as e:
            raise RuntimeError(
                f"Failed to obtain MObject attribute for plug {plug}"
            ) from e

        api_type = getattr(mobj_attr, "apiTypeStr", None)
        if not api_type:
            raise RuntimeError(
                f"Attribute object for plug {plug} has no apiTypeStr; cannot dispatch."
            )

        try:
            subclass: type[Attribute] = _API_TYPE_SUBCLASS_MAP[api_type]
        except KeyError:
            raise NotImplementedError(
                f"Unsupported attribute apiTypeStr '{api_type}'. "
                f"Known types: {sorted(_API_TYPE_SUBCLASS_MAP)}"
            )

        instance = super().__new__(subclass)
        setattr(instance, "_pre_init_plug", plug)
        return typing.cast(A, instance)

    @typing.overload
    def __init__(self, plug_or_path: str): ...

    @typing.overload
    def __init__(self, plug_or_path: OpenMaya2.MPlug): ...

    def __init__(self, plug_or_path: PlugInput):
        # _pre_init_plug is injected by Attribute.__new__ ONLY when the user called
        # Attribute(...) (base class factory dispatch). Direct subclass construction
        # (e.g. FloatAttribute(...)) bypasses that path, so pre will be None and we
        # must resolve plug_or_path here.
        pre = getattr(self, "_pre_init_plug", None)
        if pre is not None:
            self.plug = pre
            delattr(self, "_pre_init_plug")
            return

        if isinstance(plug_or_path, OpenMaya2.MPlug):
            self.plug = plug_or_path
        elif isinstance(plug_or_path, str):
            self.plug = self._plug_from_path(plug_or_path)
        else:
            raise ValueError("plug_or_path must be an MPlug or attribute path string.")

    @staticmethod
    def _plug_from_path(path: str) -> OpenMaya2.MPlug:
        """
        Resolve a string path to an MPlug.
        :param path: The full path to the attribute, e.g. '|grp|node|nodeShape.visibility'
        :return: The MPlug for the attribute.
        """
        # Split path into node path and attribute name
        if "." not in path:
            raise ValueError(
                "Attribute path must include a '.' separating node and attribute."
            )
        node_path, attr_name = path.rsplit(".", 1)
        node_dag_path = cast.get_dag_path_from_string(node_path)
        node = DagNode(node_dag_path.fullPathName())
        plug = Attribute._get_plug_from_node(node, attr_name)
        return plug

    def __str__(self):
        return self.name()

    def get(self):
        return self._get_plug_value(self.plug)

    def set(self, value):
        attr_name = self.plug.name()
        cmds.setAttr(attr_name, value)

    def node(self):
        return self._get_node_from_plug(self.plug)

    def name(self):
        raise NotImplementedError

    def connect(self, dest: "Attribute", force: bool = False, next_available=False):
        """
        Connect this attribute to another attribute.
        :param dest: The destination attribute to connect to.
        :param force: Whether to force the connection if the destination is already connected.
        :param next_available: Whether to connect to the next available index if the destination is an array attribute.
        """
        cmds.connectAttr(
            self.plug.name(),
            dest.plug.name(),
            force=force,
            nextAvailable=next_available,
        )

    def disconnect(self, dest: "Attribute", next_available=False):
        """
        Disconnect this attribute from another attribute.
        :param dest: The destination attribute to disconnect from.
        :param next_available: Whether to disconnect from the next available index if the destination is an
        array attribute.
        """
        cmds.disconnectAttr(
            self.plug.name(), dest.plug.name(), nextAvailable=next_available
        )

    @classmethod
    def _get_value(cls, node: DagNode, attr_name: str):
        plug = cls._get_plug_from_node(node, attr_name)
        return cls._get_plug_value(plug)

    @staticmethod
    def _get_plug_from_node(node: DagNode, attr_name: str) -> OpenMaya2.MPlug:
        fn_dep = node.get_mfndependency_node()
        return fn_dep.findPlug(attr_name, False)

    @staticmethod
    def _get_node_from_plug(plug: OpenMaya2.MPlug) -> Node:
        mobj = plug.node()
        if mobj.hasFn(OpenMaya2.MFn.kDagNode):
            return DagNode(OpenMaya2.MFnDagNode(mobj).fullPathName())
        else:
            name = OpenMaya2.MFnDependencyNode(mobj).name()
            return Node(name)

    @classmethod
    def _get_plug_value(cls, plug: OpenMaya2.MPlug):
        return getattr(plug, cls._getter_type)()


class FloatAttribute(Attribute):
    _getter_type = "asDouble"


class MessageAttribute(Attribute):
    _getter_type = "kMessage"

    @classmethod
    def _get_plug_value(cls, plug: OpenMaya2.MPlug):
        raise AttributeError("Message attributes do not hold data.")

    @classmethod
    def set(cls, value):
        raise AttributeError("Message attributes are not settable.")


_API_TYPE_SUBCLASS_MAP = {
    "kDoubleLinearAttribute": FloatAttribute,
    "kMessageAttribute": MessageAttribute,
}



def is_interactive_maya():
    """
    Checks if maya is running in interactive mode
    """
    import maya.cmds as cmds

    try:
        return not cmds.about(batch=True)  # noqa
    except (AttributeError, ImportError):
        return False
